-- (Tutorial by Ulf Norell - edited by Patrik Jansson.)

-- Now we're getting somewhere! Really dependent types.

module Families where -- generalization of GADTs in Haskell
-- You can import modules defined in other files.
-- More details later...
open import Basics using (⋆)
open import Nat
open import Bool
open import CurryHoward using (_∧_; _,_)

-- First an example of computing a family of datatypes
Silly : Nat -> ⋆
Silly 0             = Bool
Silly 1             = Nat
Silly (suc (suc n)) = Silly n  ∧  Silly (suc n)

-- ... and an example of a dependently typed function
silly : (n : Nat) -> Silly n
silly 0             = false
silly 1             = 1738
silly (suc (suc n)) = (silly n , silly (suc n))

test = silly 7

{- no non-const functions from ⋆ to Nat
fun : ⋆ -> Nat
fun Int = 1
-}






-- Now on to inductive families of datatypes - similar to GADTs in
-- Haskell.

data Vec (A : ⋆) : Nat -> ⋆ where
  []   : Vec A zero
  _::_ : ∀ {n : Nat} -> A -> Vec A n -> Vec A (suc n)

-- In Haskell, the same can be simulated with "type level Nat"

infixr 40 _::_

-- Some simple functions
head : {A : ⋆}{n : Nat} -> Vec A (suc n) -> A
head (x :: _) = x  -- no need for a [] case

-- Does the definition look familiar?
map : {A B : ⋆}{n : Nat} -> (A -> B) -> Vec A n -> Vec B n
map f []        = []
map f (x :: xs) = f x :: map f xs

infixr 40 _++_

_++_ : {A : ⋆}{n m : Nat} -> Vec A n -> Vec A m -> Vec A (n + m)
[]        ++ ys   =   ys
(x :: xs) ++ ys   =   x :: (xs ++ ys)

{-
_+++_ : {A : ⋆}{n m : Nat} -> Vec A n -> Vec A m -> Vec A (m + n)
[]        +++ ys   =   coerce myproof ys
(x :: xs) +++ ys   =   x :: (xs +++ ys)

coerce : (a == b) -> Vec A a == Vec A b
coerce 
-}

-- Why does this type check? Let's walk through it slowly.
-- When pattern matching on the first vector, n is instantiated.
{-
lemma : ∀ (m n : Nat) -> m + n == n + m
lemma = ?
-}

-- What happens if we make the lengths explicit?

cat : {A : ⋆}(n m : Nat) -> Vec A n -> Vec A m -> Vec A (n + m)
cat .zero    m []              ys   =   ys
cat .(suc n) m (_::_ {n} x xs) ys   =   x :: (cat n m xs ys)

-- Patterns which get instantiated by pattern matching on other stuff
-- get tagged by a dot. If you erase all the dotted things you get a
-- well-formed linear first-order pattern.

-- Let's do some other interesting families.

-- The identity type.
data _==_ {A : ⋆} : A -> A -> ⋆ where
  refl : (x : A)  ->  x == x

subst : {A : ⋆} -> (C : A -> ⋆) ->  (x y : A) ->
        x == y   ->   (C x -> C y)
subst C .x .x (refl x) cx = cx

{- Finite sets
Type             Corresponding set
Fin zero         {}
Fin (suc zero)   {fzero}
Fin 2            {fzero, fsuc fzero}
...              ...
-}
data Fin : Nat -> ⋆ where
  fzero : {n : Nat}  ->             Fin (suc n)
  fsuc  : {n : Nat}  ->  Fin n  ->  Fin (suc n)
-- Note that there are two ways of creating a number in Fin (suc n)
-- but no way to create a number in Fin zero.

-- Indexing into vectors
_!_ : {A : ⋆}{n : Nat} ->  Vec A n ->  (Fin n -> A)
[]        ! ()
(x :: xs) ! fzero  = x
(x :: xs) ! fsuc i = xs ! i

----------------
-- What's next?
-- Actually, inductive families are sufficiently fun that
-- you'll never get bored, but there's even more fun to be had.

-- Move on to: Filter.agda
